# -*- coding: utf-8 -*-

import json
import os
import requests
import shutil
import subprocess

from enum import Enum
from pkg_resources import parse_version
from werkzeug.exceptions import BadRequest, NotFound, Unauthorized

from . import crypto
from .paths import LXC_ROOT, PKG_TEMP_DIR

class Stage(Enum):
    DOWNLOAD = 1
    UNPACK = 2
    INSTALL_DEPS = 3
    INSTALL_APP = 4

class Pkg:
    def __init__(self):
        self.stage = Stage.DOWNLOAD
        self.bytes_total = 1
        self.bytes_downloaded = 0

    @property
    def percent_downloaded(self):
        # Limit the displayed percentage to 0 - 99
        return min(99, round(self.bytes_downloaded / self.bytes_total * 100))

class PkgMgr:
    def __init__(self, conf):
        self.conf = conf
        self.online_packages = {}

    def get_repo_resource(self, resource_url, stream=False):
        # Download requested repository resource
        r = requests.get('{}/{}'.format(self.conf['repo']['url'], resource_url), auth=(self.conf['repo']['user'], self.conf['repo']['pwd']), timeout=5, stream=stream)
        if r.status_code == 401:
            raise Unauthorized(r.url)
        elif r.status_code == 404:
            raise NotFound(r.url)
        elif r.status_code != 200:
            raise BadRequest(r.url)
        return r

    def fetch_online_packages(self):
        # Fetches and verifies online packages. Can raise InvalidSignature
        packages = self.get_repo_resource('packages').content
        packages_sig = self.get_repo_resource('packages.sig').content
        crypto.verify_signature(packages, packages_sig)
        self.online_packages = json.loads(packages)

    def install_app(self, app, item):
        # Main installation function. Wrapper for download, registration and install script
        self.fetch_online_packages()
        # Get all packages on which the app depends and which have not been installed yet
        deps = [d for d in self.get_install_deps(app) if d not in self.conf['packages']]
        item.bytes_total = sum(self.online_packages[d]['size'] for d in deps)
        for dep in deps:
            self.download_package(dep, item)
        for dep in deps:
            # Purge old data before unpacking to clean previous failed installation
            item.stage = Stage.UNPACK
            self.purge_package(dep)
            self.unpack_package(dep)
        for dep in deps:
            # Set stage to INSTALLING_DEPS or INSTALLING based on which package in sequence is being installed
            item.stage = Stage.INSTALL_APP if dep == deps[-1] else Stage.INSTALL_DEPS
            # Run uninstall script before installation to clean previous failed installation
            self.run_uninstall_script(dep)
            self.run_install_script(dep)
            self.register_package(dep)

    def uninstall_app(self, app):
        # Main uninstallation function. Wrapper for uninstall script, filesystem purge and unregistration
        deps = self.get_install_deps(app, False)[::-1]
        for dep in deps:
            if dep not in self.get_uninstall_deps():
                self.run_uninstall_script(dep)
                self.purge_package(dep)
                self.unregister_package(dep)

    def update_app(self, app, item):
        # Main update function.
        # TODO: Implement actual update
        uninstall_app(app)
        install_app(app, item)

    def download_package(self, name, item):
        # Download tar.xz package and verify its hash. Can raise InvalidSignature
        pkg_archive = '{}_{}-{}.tar.xz'.format(name, self.online_packages[name]['version'], self.online_packages[name]['release'])
        tmp_archive = os.path.join(PKG_TEMP_DIR, pkg_archive)
        os.makedirs(PKG_TEMP_DIR, 0o700, True)
        # If the archive already exists in temp (presumably because the previous installation was interrupted), it was already verified and can be reused
        if os.path.exists(tmp_archive):
            item.bytes_downloaded += os.path.getsize(tmp_archive)
            return
        # Download the archive
        partial_archive = '{}.partial'.format(tmp_archive)
        res = self.get_repo_resource(pkg_archive, True)
        with open(partial_archive, 'wb') as f:
            for chunk in res.iter_content(chunk_size=65536):
                if chunk:
                    item.bytes_downloaded += f.write(chunk)
        # Verify hash
        crypto.verify_hash(partial_archive, self.online_packages[name]['sha512'])
        # Remove ".partial" extension
        os.rename(partial_archive, tmp_archive)

    def unpack_package(self, name):
        # Unpack archive
        pkg_archive = '{}_{}-{}.tar.xz'.format(name, self.online_packages[name]['version'], self.online_packages[name]['release'])
        tmp_archive = os.path.join(PKG_TEMP_DIR, pkg_archive)
        subprocess.run(['tar', 'xJf', tmp_archive], cwd='/', check=True)
        os.unlink(tmp_archive)

    def purge_package(self, name):
        # Removes package and shared data from filesystem
        lxcpath = self.conf['packages'][name]['lxcpath'] if name in self.conf['packages'] else self.online_packages[name]['lxcpath']
        lxc_dir = os.path.join(LXC_ROOT, lxcpath)
        if os.path.exists(lxc_dir):
            shutil.rmtree(lxc_dir)
        srv_dir = os.path.join('/srv/', name)
        if os.path.exists(srv_dir):
            shutil.rmtree(srv_dir)
        lxc_log = '/var/log/lxc/{}.log'.format(name)
        if os.path.exists(lxc_log):
            os.unlink(lxc_log)

    def register_package(self, name):
        # Registers a package in installed packages
        metadata = self.online_packages[name].copy()
        del metadata['sha512']
        del metadata['size']
        self.conf['packages'][name] = metadata
        self.conf.save()

    def unregister_package(self, name):
        # Removes a package from installed packages
        del self.conf['packages'][name]
        self.conf.save()

    def run_install_script(self, name):
        # Runs install.sh for a package, if the script is present
        install_script = os.path.join('/srv/', name, 'install.sh')
        if os.path.exists(install_script):
            subprocess.run(install_script, check=True)

    def run_uninstall_script(self, name):
        # Runs uninstall.sh for a package, if the script is present
        uninstall_script = os.path.join('/srv/', name, 'uninstall.sh')
        if os.path.exists(uninstall_script):
            subprocess.run(uninstall_script, check=True)

    def get_install_deps(self, name, online=True):
        # Flatten dependency tree for a package while preserving the dependency order
        packages = self.online_packages if online else self.conf['packages']
        deps = packages[name]['depends'].copy()
        for dep in deps[::-1]:
            deps[:0] = [d for d in self.get_install_deps(dep, online)]
        deps = list(dict.fromkeys(deps + [name]))
        return deps

    def get_uninstall_deps(self):
        # Create reverse dependency tree for all installed packages
        deps = {}
        for name in self.conf['packages'].copy():
            for d in self.conf['packages'][name]['depends']:
                deps.setdefault(d, []).append(name)
        return deps

    def has_update(self, app):
        if not self.online_packages:
            return False
        return parse_version(self.conf['packages'][app]['version']) < parse_version(self.online_packages[app]['version'])
